@@ -1,4 +1,4 @@
-""" Demonstration of the Pmw DateTime megawidget.
+""" Demonstration of the Pmw TimeCounter megawidget.
 """
 
 # Import Pmw from the sibling directory.
@@ -34,7 +34,7 @@ class Demo:
 if __name__ == '__main__':
     root = Tkinter.Tk()
     Pmw.initialise(root, fontScheme = 'pmw1')
-    root.title('Pmw ComboBox demonstration')
+    root.title('Pmw TimeCounter demonstration')
 
     exitButton = Tkinter.Button(root, text = 'Exit', command = root.destroy)
     exitButton.pack(side = 'bottom')
