@@ -19,7 +19,6 @@ import sys
 import traceback
 import types
 import Tkinter
-from PmwUtils import forwardmethods
 
 # Constant used to indicate that an option can only be set by a call
 # to the constructor.
@@ -34,6 +33,148 @@ _OPT_FUNCTION        = 2
 
 #=============================================================================
 
+# Functions used to forward methods from a class to a component.
+
+# Fill in a flattened method resolution dictionary for a class (attributes are 
+# filtered out). Flattening honours the MI method resolution rules 
+# (depth-first search of bases in order). The dictionary has method names
+# for keys and functions for values.
+def __methodDict(cls, dict):
+
+    # the strategy is to traverse the class in the _reverse_ of the normal
+    # order, and overwrite any duplicates.
+    baseList = list(cls.__bases__)
+    baseList.reverse()
+    
+    # do bases in reverse order, so first base overrides last base
+    for super in baseList:
+	__methodDict(super, dict)
+
+    # do my methods last to override base classes
+    for key, value in cls.__dict__.items():
+	# ignore class attributes
+	if type(value) == types.FunctionType:
+	    dict[key] = value
+
+def __methods(cls):
+    # Return all method names for a class.
+
+    # Return all method names for a class (attributes are filtered
+    # out).  Base classes are searched recursively.
+
+    dict = {}
+    __methodDict(cls, dict)
+    return dict.keys()
+	
+# Function body to resolve a forwarding given the target method name and the 
+# attribute name. The resulting lambda requires only self, but will forward 
+# any other parameters.
+__stringBody = (
+    'def %(method)s(this, *args, **kw): return ' +
+    'apply(this.%(attribute)s.%(method)s, args, kw)')
+
+# Get a unique id
+__counter = 0
+def __unique():
+  global __counter
+  __counter = __counter + 1
+  return str(__counter)
+
+# Function body to resolve a forwarding given the target method name and the
+# index of the resolution function. The resulting lambda requires only self, 
+# but will forward any other parameters. The target instance is identified 
+# by invoking the resolution function.
+__funcBody = (
+    'def %(method)s(this, *args, **kw): return ' +
+    'apply(this.%(forwardFunc)s().%(method)s, args, kw)')
+
+def forwardmethods(fromClass, toClass, toPart, exclude = []):
+    # Forward all methods from one class to another.
+
+    # Forwarders will be created in fromClass to forward method
+    # invocations to toClass.  The methods to be forwarded are
+    # identified by flattening the interface of toClass, and excluding
+    # methods identified in the exclude list.  Methods already defined
+    # in fromClass, or special methods with one or more leading or
+    # trailing underscores will not be forwarded.
+
+    # For a given object of class fromClass, the corresponding toClass
+    # object is identified using toPart.  This can either be a String
+    # denoting an attribute of fromClass objects, or a function taking
+    # a fromClass object and returning a toClass object.
+
+    # Example:
+    #     class MyClass:
+    #     ...
+    #         def __init__(self):
+    #             ...
+    #             self.__target = TargetClass()
+    #             ...
+    #         def findtarget(self):
+    #             return self.__target
+    #     forwardmethods(MyClass, TargetClass, '__target', ['dangerous1', 'dangerous2'])
+    #     # ...or...
+    #     forwardmethods(MyClass, TargetClass, MyClass.findtarget, 
+    #             ['dangerous1', 'dangerous2'])
+
+    # In both cases, all TargetClass methods will be forwarded from
+    # MyClass except for dangerous1, dangerous2, special methods like
+    # __str__, and pre-existing methods like findtarget.
+
+
+    # Allow an attribute name (String) or a function to determine the instance
+    if type(toPart) != types.StringType:
+
+	# check that it is something like a function
+	if callable(toPart):
+
+	    # If a method is passed, use the function within it
+	    if hasattr(toPart, 'im_func'):
+		toPart = toPart.im_func
+		
+	    # After this is set up, forwarders in this class will use
+	    # the forwarding function. The forwarding function name is
+	    # guaranteed to be unique, so that it can't be hidden by subclasses
+	    forwardName = '__fwdfunc__' + __unique()
+	    fromClass.__dict__[forwardName] = toPart
+
+	# It's not a valid type
+	else:
+	    raise TypeError, 'toPart must be attribute name, function or method'
+
+    # get the full set of candidate methods
+    dict = {}
+    __methodDict(toClass, dict)
+
+    # discard special methods
+    for ex in dict.keys():
+	if ex[:1] == '_' or ex[-1:] == '_':
+	    del dict[ex]
+    # discard dangerous methods supplied by the caller
+    for ex in exclude:
+	if dict.has_key(ex):
+	    del dict[ex]
+    # discard methods already defined in fromClass
+    for ex in __methods(fromClass):
+	if dict.has_key(ex):
+	    del dict[ex]
+
+    for method, func in dict.items():
+	d = {'method': method, 'func': func}
+	if type(toPart) == types.StringType:
+	    execString = \
+		__stringBody % {'method' : method, 'attribute' : toPart}
+	else:
+	    execString = \
+		__funcBody % {'forwardFunc' : forwardName, 'method' : method}
+
+	exec execString in d
+
+	# this creates a method
+	fromClass.__dict__[method] = d[method]
+
+#=============================================================================
+
 class MegaArchetype:
     # Megawidget abstract root class.
 
@@ -77,7 +218,7 @@ class MegaArchetype:
 	#    - value of keyword
 	#    - a boolean indicating if the keyword has been used.
 	# A keyword is used if, during the construction of a megawidget,
-	#    - it is defined in a call to defineoptions(), or
+	#    - it is defined in a call to defineoptions() or addoptions(), or
 	#    - it references, by name, a component of the megawidget, or
 	#    - it references, by group, at least one component
 	# At the end of megawidget construction, a call is made to
@@ -142,6 +283,13 @@ class MegaArchetype:
 	    self._constructorKeywords = tmp
 	    self._optionInfo = {}
 
+	self.addoptions(optionDefs)
+
+    def addoptions(self, optionDefs):
+	# Add additional options, providing the default value and the
+	# method to call when the value is changed.  See
+	# "defineoptions" for more details
+
 	# optimisations:
 	optionInfo = self._optionInfo
 	optionInfo_has_key = optionInfo.has_key
@@ -150,8 +298,7 @@ class MegaArchetype:
 	FUNCTION = _OPT_FUNCTION
 
 	for name, default, function in optionDefs:
-	    index = string.find(name, '_')
-	    if index < 0:
+	    if '_' not in name:
 		# The option will already exist if it has been defined
 		# in a derived class.  In this case, do not override the
 		# default value of the option or the callback function
@@ -180,6 +327,9 @@ class MegaArchetype:
     def createcomponent(self, name, aliases, group, widgetClass, widgetArgs, **kw):
 	# Create a component (during construction or later).
 
+	if '_' in name:
+	    raise ValueError, 'Component name "%s" must not contain "_"' % name
+
 	if hasattr(self, '_constructorKeywords'):
 	    keywords = self._constructorKeywords
 	else:
@@ -468,28 +618,11 @@ class MegaArchetype:
 	    return widget.component(remainingComponents)
 
     def interior(self):
-	# Return the widget framing the remaining interior space.
-	# By default it returns the hull.
-
-	# Return the widget framing the interior space of the
-	# megawidget.  For the MegaWidget and MegaToplevel classes,
-	# this is the same as the 'hull' component.  When a subclass
-	# is creating components they should usually use interior() as
-	# the widget in which components should be contained. 
-	# Megawidgets (such as a toplevel window with a menu bar and
-	# status bar) which can be further subclassed should redefine
-	# interior() to return the widget in which subclasses should
-	# create their components.  The overall containing widget is
-	# always available as 'hull' (and should not be redefined).
-	
-	# This method is for use by MegaWidget or MegaToplevel
-	# subclasses.  Other classes should still access all
-	# megawidget components by name e.g.  a widget redefining
-	# interior() should create it as a component of the
-	# megawidget.
-
 	return self._hull
 
+    def hulldestroyed(self):
+	return not _hullToMegaWidget.has_key(self._hull)
+
     def __str__(self):
 	return str(self._hull)
 
@@ -577,52 +710,17 @@ class MegaArchetype:
 #=============================================================================
 
 class MegaToplevel(MegaArchetype):
-    # Toplevel megawidget base class.
-    #
-    # The MegaToplevel class inherits everything from the
-    # MegaArchetype class, and adds a Tkinter Toplevel called the
-    # 'hull' component to represent the body of the megawidget.  The
-    # window class name for the hull is set to the most-specific class
-    # name for the megawidget.  The class acts as the base class of
-    # megawidgets that are contained in their own toplevel window. 
-    # Derived classes specialise this widget by creating other widget
-    # components as children of the hull.
-    #
-    # The MegaToplevel class forwards all Tkinter Toplevel methods on
-    # to the hull component.  For example, methods such as show and
-    # activate and all the wm methods can be used with a MegaToplevel
-    # megawidget.
-    #
-    # Components
-    #   hull             exterior of the megawidget (a Tkinter Toplevel)
-    #
-    # Options
-    #   activatecommand  called whenever the toplevel is activated
-    #   title            window manager title of the toplevel window
-    #
-    # Use show/withdraw for normal use and activate/deactivate for
-    # modal dialog use.  If the window is deleted by the window
-    # manager while being shown normally, the default behaviour is to
-    # destroy the window.  If the window is deleted by the window
-    # manager while the window is active (ie:  when used as a modal
-    # dialog), the window is deactivated.  Use the userdeletefunc()
-    # and usermodaldeletefunc() methods to override these behaviours. 
-    # Do not call protocol('WM_DELETE_WINDOW', func) directly if you
-    # want to use this toplevel as a modal dialog.
 
-    # The currently active windows form a stack with the most recently
-    # activated window at the top of the stack.  All mouse and
-    # keyboard events are sent to this top window.  When it
-    # deactivates, the next window in the stack starts to receive
-    # events.  <_grabStack> is a list of tuples.  Each tuple contains
-    # the active widget and a boolean indicating whether the window
-    # was activated in global mode.
+    # <_grabStack> is a list of tuples.  Each tuple contains the
+    # active widget and a boolean indicating whether the window was
+    # activated in global mode.
     _grabStack = []
 
     def __init__(self, parent = None, **kw):
 	# Define the options for this megawidget.
 	optiondefs = (
 	    ('activatecommand',  None,  None),
+	    ('deactivatecommand',  None,  None),
 	    ('title',            None,  self._settitle),
 	    ('hull_class',       self.__class__.__name__,  None),
 	)
@@ -688,15 +786,48 @@ class MegaToplevel(MegaArchetype):
 	    self.tkraise()
 	else:
 	    if self._firstShowing:
+		# Just let the window manager determine the window
+		# position for the first time.
 	        self._firstShowing = 0
 	    else:
+		# Position the window at the same place it was last time.
 		geometry = self.geometry()
 		index = string.find(geometry, '+')
 		if index >= 0:
 		    self.geometry(geometry[index:])
 	    self.deiconify()
 
-    def activate(self, globalMode=0, master=None):
+    def _centreonscreen(self):
+	# Centre the window on the screen.  (Actually halfway across
+	# and one third down.)
+
+	self.update_idletasks()
+
+	# I'm not sure what the winfo_vroot[xy] stuff does, but tk_dialog
+	# does it, so...
+	#x = (self.winfo_screenwidth() - self.winfo_reqwidth()) / 2
+	#y = (self.winfo_screenheight() - self.winfo_reqheight()) / 3
+
+	x = (self.winfo_screenwidth() - self.winfo_reqwidth()) / 2 \
+		- self.winfo_vrootx()
+	y = (self.winfo_screenheight() - self.winfo_reqheight()) / 3 \
+		- self.winfo_vrooty()
+	if x < 0:
+	    x = 0
+	if y < 0:
+	    y = 0
+	self.geometry('+%s+%s' % (x, y))
+
+    def _sameposition(self):
+	# Position the window at the same place it was last time.
+
+	geometry = self.geometry()
+	index = string.find(geometry, '+')
+	if index >= 0:
+	    self.geometry(geometry[index:])
+
+    def activate(self, globalMode=0, master=None,
+	    geometry = 'centerscreenfirst'):
 	if self.state() == 'normal':
 	    self.withdraw()
 	if self._active == 'yes':
@@ -720,27 +851,34 @@ class MegaToplevel(MegaArchetype):
 
 	self._active = 'waiting'
 
-	# Centre the window on the screen. (Actually halfway across and
-	# one third down.)
-	self.update_idletasks()
+	if geometry == 'centerscreenalways':
+	    self._centreonscreen()
+	elif geometry == 'centerscreenfirst':
+	    if self._firstShowing:
+		# Centre the window the first time it is displayed.
+		self._centreonscreen()
+	    else:
+		# Position the window at the same place it was last time.
+		self._sameposition()
+	elif geometry[:5] == 'first':
+	    if self._firstShowing:
+		self.geometry(geometry[5:])
+	    else:
+		# Position the window at the same place it was last time.
+		self._sameposition()
+	elif geometry is not None:
+	    self.geometry(geometry)
 
-	# I'm not sure what the winfo_vroot[xy] stuff does, but tk_dialog
-	# does it, so...
-	#x = (self.winfo_screenwidth() - self.winfo_reqwidth()) / 2
-	#y = (self.winfo_screenheight() - self.winfo_reqheight()) / 3
-
-	x = (self.winfo_screenwidth() - self.winfo_reqwidth()) / 2 \
-		- self.winfo_vrootx()
-	y = (self.winfo_screenheight() - self.winfo_reqheight()) / 3 \
-		- self.winfo_vrooty()
-	if x < 0:
-	    x = 0
-	if y < 0:
-	    y = 0
-	self.geometry('+%s+%s' % (x, y))
+	self._firstShowing = 0
 	self.deiconify()
 
 	self.wait_visibility()
+
+	if self._active == 'no':
+	    # The deactivate() method was called during the call to
+	    # wait_visibility() (perhaps from a timer).
+	    return self._result
+
 	self._active = 'yes'
 
 	while 1:
@@ -751,7 +889,6 @@ class MegaToplevel(MegaArchetype):
 		    self.grab_set()
 		break
 	    except Tkinter.TclError:
-		sys.exc_traceback = None   # Clean up object references
 		# Another application has grab.  Keep trying until
 		# grab can succeed.
 		self.after(100)
@@ -806,7 +943,7 @@ class MegaToplevel(MegaArchetype):
     #     }
 
     def deactivate(self, result=None):
-	if not self.active():
+	if self._active == 'no':
 	    return
 	self._active = 'no'
 
@@ -817,7 +954,6 @@ class MegaToplevel(MegaArchetype):
 	  else:
 	    MegaToplevel._grabStack[-1][0].deactivate()
 
-
 	# Clean up this window.
 	hidebusycursor()
 	self.withdraw()
@@ -827,10 +963,21 @@ class MegaToplevel(MegaArchetype):
 	del MegaToplevel._grabStack[-1]
 	if len(MegaToplevel._grabStack) > 0:
 	    widget, globalMode = MegaToplevel._grabStack[-1]
-	    if globalMode:
-		widget.grab_set_global()
-	    else:
-		widget.grab_set()
+	    while 1:
+		try:
+		    if globalMode:
+			widget.grab_set_global()
+		    else:
+			widget.grab_set()
+		    break
+		except Tkinter.TclError:
+		    # Another application has grab.  Keep trying until
+		    # grab can succeed.
+		    self.after(100)
+
+	command = self['deactivatecommand']
+	if callable(command):
+	    command()
 
 	self._result = result
 	self._wait.set(1)
@@ -843,29 +990,6 @@ forwardmethods(MegaToplevel, Tkinter.Toplevel, '_hull')
 #=============================================================================
 
 class MegaWidget(MegaArchetype):
-    # Frame megawidget base class.
-    #
-    # The MegaWidget class inherits everything from the MegaArchetype
-    # class, and adds a Tkinter Frame called the 'hull' component to
-    # represent the body of the megawidget.  The window class name for
-    # the hull is set to the most-specific class name for the
-    # megawidget.  The class acts as the base class of megawidgets
-    # that are not contained in their own toplevel window.  Derived
-    # classes specialize this widget by creating other widget
-    # components as children of the hull.
-    #
-    # The MegaWidget class forwards all Tkinter Frame methods on to
-    # the hull component.  For example, methods such as pack and
-    # configure and all the winfo_ methods can be used with a
-    # MegaWidget.
-    #
-    # Components
-    #   hull          exterior of the megawidget (a Tkinter Frame)
-
-    # Options
-    #   background    forwarded to hull component
-    #   cursor        forwarded to hull component
-
     def __init__(self, parent = None, **kw):
 	# Define the options for this megawidget.
 	optiondefs = (
@@ -950,7 +1074,6 @@ def hidebusycursor():
 			# Previous focus widget has been deleted. Set focus
 			# to toplevel window instead (can't leave focus on
 			# busy window).
-			sys.exc_traceback = None   # Clean up object references
 			window.focus_set()
 		    if window._w == '.':
 			busyWindow = '._Busy'
@@ -972,7 +1095,6 @@ def clearbusycursor():
 		    # Previous focus widget has been deleted. Set focus
 		    # to toplevel window instead (can't leave focus on
 		    # busy window).
-		    sys.exc_traceback = None   # Clean up object references
 		    window.focus_set()
 
 def busycallback(command, updateFunction = None):
@@ -983,6 +1105,9 @@ def busycallback(command, updateFunction = None):
     wrapper = _BusyWrapper(command, updateFunction)
     return wrapper.callback
 
+_errorReportFile = None
+_errorWindow = None
+
 def reporterrorstofile(file = None):
     global _errorReportFile
     _errorReportFile = file
@@ -990,11 +1115,13 @@ def reporterrorstofile(file = None):
 def displayerror(text):
     global _errorWindow
 
-    if _errorWindow is None:
-	# The error window has not yet been created.
-	_errorWindow = _ErrorWindow()
-
-    _errorWindow.showerror(text)
+    if _errorReportFile is not None:
+	_errorReportFile.write(text + '\n')
+    else:
+	if _errorWindow is None:
+	    # The error window has not yet been created.
+	    _errorWindow = _ErrorWindow()
+	_errorWindow.showerror(text)
 
 def initialise(root = None, size = None, fontScheme = None, useTkOptionDb = 0):
 
@@ -1036,7 +1163,7 @@ def initialise(root = None, size = None, fontScheme = None, useTkOptionDb = 0):
     # Set the base font size for the application and set the
     # Tk option database font resources.
     import PmwLogicalFont
-    PmwLogicalFont.initialise(root, size, fontScheme)
+    PmwLogicalFont._font_initialise(root, size, fontScheme)
 
     return root
 
@@ -1150,6 +1277,8 @@ def __addRootToToplevelBusyCount():
     # last minute.
 
     root = Tkinter._default_root
+    if root == None:
+	root = Tkinter.Tk()
     if not _toplevelBusyInfo.has_key(root):
 	_toplevelBusyInfo[root] = [0, None]
 
@@ -1186,7 +1315,6 @@ def __TkinterToplevelDestroy(tkWidget):
 	  mega.destroy()
       except:
 	  _reporterror(mega.destroy, ())
-	  sys.exc_traceback = None   # Clean up object references
     else:
       del _toplevelBusyInfo[tkWidget]
       Tkinter.Widget.destroy(tkWidget)
@@ -1198,7 +1326,6 @@ def __TkinterFrameDestroy(tkWidget):
 	  mega.destroy()
       except:
 	  _reporterror(mega.destroy, ())
-	  sys.exc_traceback = None   # Clean up object references
     else:
       Tkinter.Widget.destroy(tkWidget)
 
@@ -1224,7 +1351,6 @@ class __TkinterCallWrapper:
 	    raise SystemExit, msg
 	except:
 	    _reporterror(self.func, args)
-	    sys.exc_traceback = None   # Clean up object references
 
 _eventTypeToName = {
     2 : 'KeyPress',
@@ -1263,9 +1389,8 @@ _eventTypeToName = {
 }
 
 def _reporterror(func, args):
-    # Save current exception values, just in case a new one occurs.
-    exc_type, exc_value, exc_traceback = \
-      sys.exc_type, sys.exc_value, sys.exc_traceback
+    # Fetch current exception values.
+    exc_type, exc_value, exc_traceback = sys.exc_info()
 
     # Give basic information about the callback exception.
     if type(exc_type) == types.ClassType:
@@ -1303,20 +1428,12 @@ def _reporterror(func, args):
 	    msg = msg + '    %s: %s\n' % (key, args[0].__dict__[key])
 
     clearbusycursor()
-    if _errorReportFile is not None:
-	_errorReportFile.write(msg + '\n')
-    else:
-	try:
-	    displayerror(msg)
-	except:
-	    exc_type, exc_value, exc_traceback = \
-	      sys.exc_type, sys.exc_value, sys.exc_traceback
-	    print 'Failed to display error window.'
-	    print 'Original error was:'
-	    print msg
-
-_errorReportFile = None
-_errorWindow = None
+    try:
+	displayerror(msg)
+    except:
+	print 'Failed to display error window.'
+	print 'Original error was:'
+	print msg
 
 class _ErrorWindow:
     def __init__(self):
